
import org.deeplearning4j.rl4j.space.ObservationSpace;
import org.deeplearning4j.rl4j.space.ArrayObservationSpace;
import org.deeplearning4j.rl4j.space.DiscreteSpace;
import org.deeplearning4j.rl4j.space.Encodable;
import org.deeplearning4j.rl4j.mdp.MDP;
import org.deeplearning4j.gym.StepReply;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.factory.Nd4j;
import com.github.sh0nk.matplotlib4j.Plot;
import com.github.sh0nk.matplotlib4j.PythonConfig;
import com.github.sh0nk.matplotlib4j.PythonExecutionException;
import java.io.IOException;
import java.util.Random;
import java.util.ArrayList;
import java.util.List;

// MDPインターフェースの全てのメソッドを実装したMoveMDPクラスを定義
public class MoveMDP implements MDP<Encodable, Integer, DiscreteSpace> {
	double pos;						// 位置 -1.0〜1.0
	double vel;						// 速度 -1.0〜1.0
	boolean done;					// episode終了フラッグ
	int stepCnt;					// test時に300回で停止させるためのstepカウンタ
	int episCnt = 0;				//	何回目のepisodeかをカウントするカウンタ
	boolean omtFlg = false;			// posLogが初期値のみの場合に次のposLogを表示するためのフラッグ
	boolean modeFlg = false;		// train: false, test: true
	// 全stepにおけるposをepidode毎に記録するリストを生成
	List<Double> posLog = new ArrayList<Double>();
	// testモードで異なるreset()をまたいでaddしたデータをplotするため、ここでインスタンスを宣言
	// OSによってcreateの引数が異なるため、createはコンストラクタ内で行う
	Plot plt;
	// コンストラクタでインスタンス生成時の引数を受取り、pltをcreateする
	MoveMDP(boolean modeFlg){
		this.modeFlg = modeFlg;	
		// OSがWindowsかどうかを判別して、python実行ファイルのpathを指定する
		boolean isWindows = System.getProperty("os.name").toLowerCase().startsWith("windows");
		String pyPath;
		if(isWindows) {
			pyPath = "C:\\Users\\kenichi\\anaconda3\\python.exe";
		} else {
			pyPath = "/home/kenichi/anaconda3/bin/python3";
		}
		plt = Plot.create(PythonConfig.pythonBinPathConfig(pyPath));
	}
    // 観測空間と行動空間のサイズを規定: ArrayObservationSpaceではINDArray(deeplearning4jにあるNumPyの
    // 代用)が使われており、"Nd4j.create(new int[]{2})"で要素"0"が2個の1次元INDArrayを生成
    // (ここで指定されているintは{}の中の数値がintという意味で、配列要素がintという意味ではない)
	ObservationSpace<Encodable> observationSpace = new ArrayObservationSpace<>(new int[]{2});
	DiscreteSpace actionSpace = new DiscreteSpace(3);
	
	public ObservationSpace<Encodable> getObservationSpace() {
		return observationSpace;
	}
	
	public DiscreteSpace getActionSpace() {
		return actionSpace;
	}
	
	public Encodable getObservation() {
		return new Encodable() {
			// 位置と速度を配列に変換して返す
			public double[] toArray() {
				double[] state = new double[2];
				state[0] = pos;
				state[1] = vel;
				return state;
			}
			// 配列からINDArrayに変換して返す
			public INDArray getData() {
				double[] state = this.toArray();
				INDArray stateIND = Nd4j.create(state);
				return stateIND;
			}
			/* DQNPOlicyが継承するPolicyの中に、
			 * if (obs.isSkipped()) { action = lastAction; }
			 * else { action = nextAction(obs); }	
    		 * lastAction = action;
    		 * と書かれており、isSkipped()がtrueだとactionが更新されない */
			public boolean isSkipped() {
		    	return false;
		    }
		    // よく分からないが、実装しないとエラーになる
		    public Encodable dup() {
		    	return null;
		    }
		};
    }
    
	public Encodable reset() {
		// episode終了フラッグのリセット
		done = false;
		// stepカウンタのリセット
		stepCnt = 0;
		// エピソード10回分のグラフを重ねて表示
		plt.plot().add(posLog);
		if(episCnt % 10 == 0 && episCnt != 0) {
			plt.ylim(-1., 1.);
			plt.title("Step vs Position for 10 Episodes until " + episCnt);
			plt.xlabel("STEPS");
			plt.ylabel("POSITION");
			try {
				plt.show();
			} catch (IOException|PythonExecutionException e) {
				System.out.println(e);
			}
		}
		// posLogリストをクリア
		posLog.clear();
		// episodeカウンタを加算
		episCnt++;
		// 初期位置は-1.0〜1.0のランダム、速度は0.0
		Random rand = new Random();
		pos = rand.nextDouble() * 2. - 1.;
      vel = 0.;
      return getObservation();
	}
	
	public void close() {
		plt.close();
	}
	
	// 各stepごとに呼ばれるメソッド
	// actionを受け取り、次のstateとreward、episodeが終了したかどうかを返す
	public StepReply<Encodable> step(Integer action) {
		stepCnt++;
		double reward;
		double dt = 0.1;			// Δtのこと
        // 3種のaction(0, 1, 2)を加速度(acc)に変換
		double acc = ((int)action - 1.) * 0.1;
		vel += acc * dt;
        // 速度の最小値を-1.0、最大値を1.0に設定する
		vel = Math.max(-1.0,  Math.min(vel, 1.0));
		pos += vel * dt;
        // 位置の最小値を-1.0、最大値を1.0に設定する
		pos = Math.max(-1.0,  Math.min(pos, 1.0));
			// posLogリストにposを追記
		posLog.add(pos);
		
		// 位置と速度の絶対値が0.1より小さくなったらepisodeを終了
		done = (Math.abs(pos) < 0.1) && (Math.abs(vel) < 0.1);
      if(done) {
    		// 終了したときに正の報酬
         reward = 1.0;
      }else{
        	// 時間経過ごとに負の報酬
       		// ゴールへの距離が近くなるほど絶対値を減らしておくと、学習が早く進む
    	  	reward = -0.01 * Math.abs(pos);
      	}
    	if(stepCnt > 300) {
    		done = true;
    	}
		return new StepReply(getObservation(), reward, isDone(), null);
	}
	
	public boolean isDone() {
		return done;
	}
	
	// DQNでは必要ないが、実装しないとエラーになる
	public MDP<Encodable, Integer, DiscreteSpace> newInstance(){
		return null;
	}
}
